import theano.tensor as T
from theano import dot, function, map, shared
from math import exp, e
import numpy as np

# No. 1: Using map
x = T.vector('x')
W = T.matrix('W')
f = lambda a, b: e**dot(a, b)
results, updates = map(f, sequences=[W], non_sequences=x)

f_np = lambda a, b: e**np.dot(a, b)
v = np.array([0,1,2])
print(f_np(v, v))

calculate_results = function([W, x], outputs=results, allow_input_downcast=True)
test_W = np.array([[0,1,2],[0,1,2]])
test_x = np.array([0,1,2])
print(calculate_results(test_W, test_x))

# No. 2: The softmax function
x = T.vector('x')
y = T.nnet.softmax(x)
smax = function([x], y, allow_input_downcast=True)
print(smax(np.array([0,1,2])))
elements = [0,1,2]
total = sum([exp(el) for el in elements])
print([exp(el)/total for el in elements])



from math import exp, e
import pymc3
from pysterior import linear_regression
import numpy as np
from theano import tensor as T
from theano import map, function

class AbstractLogisticRegression(linear_regression._AbstractModel):
    def get_predictive_posterior_samples(self, x):
        raise NotImplementedError

    def predict(self, x):
        "Approximates the expected value of the output variable."
        s = self.get_predictive_posterior_samples(x)
        return sum(s) / len(s)

class BinaryLogisticRegressionModel(AbstractLogisticRegression):
    def _build_model(self, X, y):
        lr_model = pymc3.Model()
        with lr_model:
            X = pymc3.Normal(name='X', mu=1, sd=2, observed=X)
            precision = pymc3.Uniform(name='precision')
            w = pymc3.Normal(name='w', mu=0, sd=1.0 / precision, shape=self.input_data_dimension)
            bias_precision = pymc3.Uniform(name='bias_precision')
            bias = pymc3.Normal(name='bias', mu=0, sd=1.0 / bias_precision)
            p = 1.0/(1.0 + e**(-w.dot(X.T) - bias))
            y_obs = pymc3.Bernoulli(p=p, name='y_obs', observed=y)

        return lr_model

    def get_predictive_posterior_samples(self, x):
        "Obtain a sample of the output variable's distribution by running the sample variable values through the model."
        predictive_posterior_samples = []
        for bias, w in zip(self.samples['bias'], self.samples['w']):
            predictive_posterior_samples.append(1.0/(1.0 + e**(-bias - np.dot(x, w))))
        return predictive_posterior_samples

class LogisticRegressionModel(AbstractLogisticRegression):
    def _build_model(self, X, y):
        output_dimension = max(y)+1
        lr_model = pymc3.Model()
        with lr_model:
            X = pymc3.Normal(name='X', observed=X)
            precision = 0.00001
            weights = pymc3.Normal(name='weights',
                                   tau=precision,
                                   shape=(output_dimension, self.input_data_dimension),
                                   testval=T.zeros((output_dimension, self.input_data_dimension)))
            bias_precision = 0.00001
            biases = pymc3.Normal(name='biases', mu=0, tau=bias_precision, shape=output_dimension)
            dots = weights.dot(X.T) # No bias!
            p = T.nnet.softmax(dots + biases.T)
            y_obs = pymc3.Categorical(p=p, name='y_obs', observed=y)

        return lr_model, weights, biases

    def fit(self, X, y, sampling_iterations):
        X = self._force_shape(X)
        self.input_data_dimension = len(X[0])
        model, w, b = self._build_model(X, y)
        with model:
            self.map_estimate = pymc3.find_MAP(model=model, vars=[w, b])
            step = pymc3.NUTS(scaling=self.map_estimate)
            trace = pymc3.sample(sampling_iterations, step, start=self.map_estimate)
        self.samples = trace

    def get_predictive_posterior_samples(self, x):
        "Obtain a sample of the output variable's distribution by running the sample variable values through the model."
        predictive_posterior_samples = []
        for bias, w in zip(self.samples['bias'], self.samples['w']):
            predictive_posterior_samples.append(1.0/(1.0 + e**(-bias - np.dot(x, w))))
        return predictive_posterior_samples


import math
import random
import numpy as np

class PosteriorSample(object):
    def __init__(self):
        self.samples = []
        self.average = None

    def get_samples(self):
        return self.samples

    def add_sample(self, new_sample):
        self.samples.append(new_sample)
        self._update_average(new_sample)

    def _update_average(self, new_sample):
        if self.average is None:
            self.average = new_sample
        else:
            self.average = self.average + ((new_sample - self.average)/len(self.samples))

    def get_mean(self):
        return self.average

    def get_median(self):
        #TODO: Make this an online calculation
        center = (len(self.samples)-1)/2
        left, right = math.floor(center), math.ceil(center)
        sorted_data = sorted(self.samples)
        return (sorted_data[left] + sorted_data[right])/2.0

    def get_random(self, n=1):
        return np.array([random.choice(self.samples) for i in range(n)])



from theano import tensor as T
import theano
from theano.tensor import nlinalg
from collections import namedtuple
import numpy as np
X = 'X'
MU = 'Mu'

FunctionSpec = namedtuple('FunctionSpec', ['variables', 'output_expression'])

Energy = namedtuple('Energy', ['eval', 'gradient'])

def get_normal_spec():
    X,mu,sigma = [T.vector('X'), T.vector('Mu'), T.matrix('Sigma')]
    GaussianDensitySpec = FunctionSpec(variables=[X, mu, sigma],
                                       output_expression = -0.5*T.dot(T.dot((X-mu).T, nlinalg.matrix_inverse(sigma)), (X-mu)))
    return GaussianDensitySpec

def get_bivariate_normal_spec():
    X1,X2,mu,sigma = [T.scalar('X1'),T.scalar('X2'), T.vector('mu'), T.matrix('sigma')]
    GaussianDensitySpec = FunctionSpec(variables=[X1, X2, mu, sigma],
                                       output_expression = -0.5*T.dot(T.dot((T.concatenate([X1.dimshuffle('x'),X2.dimshuffle('x')])-mu).T,
                                                                            nlinalg.matrix_inverse(sigma)),
                                                                      (T.concatenate([X1.dimshuffle('x'),X2.dimshuffle('x')])-mu)))
    return GaussianDensitySpec


class PartiallyDifferentiableFunctionFactory(object):
    def __init__(self, func_spec):
        self.f = theano.function(func_spec.variables,
                                 func_spec.output_expression,
                                 allow_input_downcast=True)
        self.variables = func_spec.variables
        self.var_lookup = {v.name:v for v in func_spec.variables}
        self.output_expression = func_spec.output_expression

    def get_partial_diff(self, differentiable_var_name):
        diff_var = self.var_lookup[differentiable_var_name]
        grad = theano.function(self.variables,
                               theano.grad(self.output_expression,
                                           diff_var),
                               allow_input_downcast=True)
        return self.f, grad

def build_kwarg_closure(f, bound_kwargs):
    def partial_fxn(**kwargs):
        full_kwargs = {}
        full_kwargs.update(bound_kwargs)
        full_kwargs.update(kwargs)
        return f(**full_kwargs)
    return partial_fxn

def build_arg_closure(f, bound_kwargs, unbound_arg_name):
    def partial_fxn(arg):
        full_kwargs = {}
        full_kwargs.update(bound_kwargs)
        full_kwargs.update({unbound_arg_name: arg})
        return f(**full_kwargs)
    return partial_fxn


import numpy as np
import pymc3

class _AbstractModel(object):
    def fit(self, X, y, sampling_iterations):
        X = self._force_shape(X)
        self.input_data_dimension = len(X[0])
        model = self._build_model(X, y)
        with model:
            self.map_estimate = pymc3.find_MAP(model=model)
            step = pymc3.NUTS(scaling=self.map_estimate)
            trace = pymc3.sample(sampling_iterations, step, start=self.map_estimate)
        self.samples = trace

    def _build_model(self, X, y):
        raise NotImplementedError

    def _force_shape(self, X):
        shape = np.shape(X)
        if len(shape) == 1:
            return np.reshape(X, (shape[0], 1))
        return X

    def get_map_estimate(self):
        return self.map_estimate

    def get_samples(self):
        return self.samples

class _AbstractLinearRegression(_AbstractModel):
    def get_predictive_posterior_samples(self, x):
        "Obtain a sample of the output variable's distribution by running the sample variable values through the model."
        predictive_posterior_samples = []
        for alpha, beta in zip(self.samples['alpha'], self.samples['beta']):
            predictive_posterior_samples.append(alpha + np.dot(x, beta))
        return predictive_posterior_samples

    def predict(self, x):
        "Approximates the expected value of the output variable."
        s = self.get_predictive_posterior_samples(x)
        return sum(s) / len(s)

    def _build_model(self, X, y):
        lr_model = pymc3.Model()

        with lr_model:
            alpha = self._get_alpha()
            beta = self._get_beta()
            sigma = self._get_sigma()
            X = pymc3.Normal(name='X', mu=1, sd=2, observed=X)
            mu = alpha + beta.dot(X.T)
            Y_obs = pymc3.Normal(name='Y_obs', mu=mu, sd=sigma, observed=y)

        return lr_model

    def _get_sigma(self):
        noise_precision = pymc3.Uniform(name='noise_precision')
        sigma = pymc3.HalfNormal(name='sigma', sd=1.0 / noise_precision)
        return sigma

    def _get_alpha(self):
        raise NotImplementedError

    def _get_beta(self):
        raise NotImplementedError

class LinearRegression(_AbstractLinearRegression):
    def _get_alpha(self):
        alpha_precision = pymc3.Uniform(name='alpha_precision')
        alpha = pymc3.Normal(name='alpha', mu=0, sd=1.0 / alpha_precision)
        return alpha

    def _get_beta(self):
        precision = pymc3.Uniform(name='precision')
        beta = pymc3.Normal(name='beta', mu=0, sd=1.0 / precision, shape=self.input_data_dimension)
        return beta

class RidgeRegression(_AbstractLinearRegression):
    def __init__(self, weight_prior_sdev):
        self.weight_prior_sdev = weight_prior_sdev

    def _get_alpha(self):
        alpha = pymc3.Normal(name='alpha', mu=0, sd=self.weight_prior_sdev)
        return alpha

    def _get_beta(self):
        precision = pymc3.Uniform(name='precision')
        beta = pymc3.Normal(name='beta', mu=0, sd=1.0 / precision, shape=self.input_data_dimension)
        return beta

class LassoRegression(_AbstractLinearRegression):
    def __init__(self, weight_prior_scale):
        self.weight_prior_scale = weight_prior_scale

    def _get_alpha(self):
        alpha = pymc3.Laplace('alpha', 0, self.weight_prior_scale)
        return alpha

    def _get_beta(self):
        beta = pymc3.Laplace('beta', 0, self.weight_prior_scale, shape=self.input_data_dimension)
        return beta

class RobustLinearRegression(LinearRegression):
    def _get_sigma(self):
        noise_precision = pymc3.Uniform(name='noise_precision')
        sigma = pymc3.HalfCauchy(name='sigma', beta=1.0 / noise_precision)
        return sigma


import unittest
from pysterior import classification
import numpy as np

class LogisticRegressionTest(unittest.TestCase):
    def test_binary_log_reg(self):
        X = [[1, 2], [0, 1], [1, 0], [2, 0]]
        y = [1, 1,  0, 0]
        lr = classification.BinaryLogisticRegressionModel()
        lr.fit(X, y, 1000)
        for x in X:
            print(lr.get_predictive_posterior_samples(x))
            print(lr.predict(x))

    def test_log_reg(self):
        X = np.array([[1, 1], [-1, 1], [-1, 1], [-1, -1]])
        y = [0, 1,  0, 1]
        lr = classification.LogisticRegressionModel()
        lr.fit(X, y, 1000)
        for x in X:
            print(lr.get_predictive_posterior_samples(x))
            print(lr.predict(x))


import random
import math
import numpy as np

class LeapfrogIntegrator(object):
    def __init__(self, target_energy_gradient):
        self.target_energy_gradient = target_energy_gradient

    def _leapfrog_step(self, value, momentum, step_size):
        half_step_momentum = momentum + (step_size*0.5*self.target_energy_gradient(value))
        step_value = value + (step_size*half_step_momentum)
        step_momentum = half_step_momentum + (step_size*0.5*self.target_energy_gradient(step_value))
        return step_value, step_momentum

    def run_leapfrog(self, current_value, current_momentum, num_steps, step_size):
        value, momentum = current_value, current_momentum
        for i in range(num_steps):
            value,momentum = self._leapfrog_step(value, momentum, step_size)
        return value, momentum


class HamiltonianMCProcess(object):
    def _sample_momentum(self, dimension):
        if dimension == 0:
            return np.random.normal()
        else:
            return np.random.normal(size=dimension)

    def _get_dimension(self, p):
        try: #TODO: Remove try/except
            return len(p)
        except TypeError:
            return 0

    def _get_probability(self, energy, p, r):
        return math.exp(energy.eval(p) - (0.5 * np.dot(r,r)))

    def _get_log_probability(self, energy, p, r):
        return energy.eval(p) - (0.5 * np.dot(r,r))


class RobbinsMonroEpsilonEstimator(HamiltonianMCProcess):
    def estimate_epsilon(self, energy, initial_point, iterations=10):
        epsilon = self._select_heuristic_epsilon(energy, initial_point, iterations)
        print('Selected epsilon = ', epsilon)
        return epsilon

    def _select_heuristic_epsilon(self, energy, initial_point, iterations):
        TARGET_ACCEPTANCE = 0.65
        epsilon = 1.0
        rate_initial_value = 1.0
        for i in range(0,iterations):
            acceptance_rate = self._sample_acceptance_rate(energy, epsilon, initial_point, 100)
            learning_rate = rate_initial_value / (i+1)
            epsilon = epsilon - learning_rate*(TARGET_ACCEPTANCE - acceptance_rate)
        return epsilon

    def _sample_acceptance_rate(self, energy, epsilon, initial_point, iterations):
        leapfrog = LeapfrogIntegrator(energy.gradient)
        accepted = 0
        total = 0
        dimension = self._get_dimension(initial_point)
        for i in range(iterations):
            momentum = self._sample_momentum(dimension)
            next_point, next_momentum = leapfrog.run_leapfrog(initial_point, momentum, 10, epsilon)
            log_acc_prob = min(math.log(1.0),
                               self._get_log_probability(energy, next_point, next_momentum) - self._get_log_probability(
                                   energy, initial_point, momentum))
            if math.log(random.random()) < log_acc_prob:
                accepted += 1
            total += 1
        acceptance_rate = 1.0 * accepted / total
        return acceptance_rate

class NUTS(HamiltonianMCProcess):
    def nuts_with_initial_epsilon(self, initial_point, energy, iterations, burn_in=0):
        epsilon = RobbinsMonroEpsilonEstimator().estimate_epsilon(energy, initial_point)
        return self.nuts_with_fixed_epsilon(initial_point, energy, epsilon, iterations, burn_in)

    def nuts_with_fixed_epsilon(self, initial_point, energy, epsilon, iterations, burn_in=0):
        dimension = self._get_dimension(initial_point)
        samples = []
        current_sample = initial_point
        for i in range(iterations+burn_in):
            momentum = self._sample_momentum(dimension)
            slice_edge = random.uniform(0, self._get_probability(energy, current_sample, momentum))
            forward = back = current_sample
            forward_momentum = back_momentum = momentum
            next_sample = current_sample
            n = 1
            no_u_turn = True
            j = 0
            while no_u_turn:
                direction = random.choice([-1, 1])
                if direction == 1:
                    _, _, forward, forward_momentum, candidate_point, candidate_n, candidate_no_u_turn = self._build_tree(forward, forward_momentum, slice_edge, direction, j, epsilon, energy)
                else:
                    back, back_momentum, _, _, candidate_point, candidate_n, candidate_no_u_turn = self._build_tree(back, back_momentum, slice_edge, direction, j, epsilon, energy)
                if candidate_no_u_turn:
                    prob = min(1.0, 1.0*candidate_n/n)
                    if random.random() < prob:
                        next_sample = candidate_point
                n = n + candidate_n
                no_u_turn = candidate_no_u_turn and (np.dot((forward - back) , back_momentum) > 0) and (np.dot((forward - back) , forward_momentum) > 0)
                j += 1
            if i >= burn_in:
                samples.append(next_sample)
            current_sample = next_sample
        return samples

    def _build_tree(self, point, momentum, slice_edge, direction, j, epsilon, energy):
        leapfrog = LeapfrogIntegrator(energy.gradient)
        if j == 0:
            p, r = leapfrog.run_leapfrog(point, momentum, 1, direction*epsilon)
            if slice_edge > 0: #TODO: This is an ugly hack; find a numerically stable solution or hide it when we refactor
                candidate_n = self.I(slice_edge < self._get_probability(energy, p, r))
                candidate_no_u_turn = (self._get_probability(energy, p, r) > math.log(slice_edge) - 1000)
            else:
                candidate_n = 1
                candidate_no_u_turn = True
            return p, r, p, r, p, candidate_n, candidate_no_u_turn
        else:
            back, back_momentum, forward, forward_momentum, candidate_point, candidate_n, candidate_no_u_turn = self._build_tree(point, momentum, slice_edge, direction, j-1, epsilon, energy)
            if candidate_no_u_turn:
                if direction == 1:
                    _, _, forward, forward_momentum, candidate_point_2, candidate_n_2, candidate_2_no_u_turn = self._build_tree(forward, forward_momentum, slice_edge, direction, j-1, epsilon, energy)
                else:
                    back, back_momentum, _, _, candidate_point_2, candidate_n_2, candidate_2_no_u_turn = self._build_tree(back, back_momentum, slice_edge, direction, j-1, epsilon, energy)
                if candidate_n_2 > 0 and random.random() < (candidate_n_2 / (candidate_n_2 + candidate_n)):
                    candidate_point = candidate_point_2
                candidate_no_u_turn = candidate_2_no_u_turn and (np.dot((forward - back) , back_momentum) > 0) and (np.dot((forward - back) , forward_momentum) > 0)
                candidate_n = candidate_n + candidate_n_2
            return back, back_momentum, forward, forward_momentum, candidate_point, candidate_n, candidate_no_u_turn


    def I(self, statement):
        if statement == True:
            return 1
        else:
            return 0


import pymc3
import random
from matplotlib import pyplot as plt
import numpy as np

def test_bernoulli():
    data = [random.randint(0,1) for i in range(200)]

    model = pymc3.Model()

    with model:
        p = pymc3.Uniform(lower=0,upper=1, name='p')
        X = pymc3.Bernoulli(p=p, name='X', observed=data)

        start = pymc3.find_MAP()

        # instantiate sampler
        step = pymc3.NUTS(scaling=start)

        # draw 500 posterior samples
        trace = pymc3.sample(10000, step, start=start)

    pymc3.traceplot(trace)
    plt.show()

def test_categorical():
    k = 3
    ndata = 5000

    v = np.random.randint(0, k, ndata)

    with pymc3.Model() as model:
        p = pymc3.Dirichlet(name='p', a=np.array([1., 1., 1.]), shape=k)
        category = pymc3.Categorical(name='category',
                                     p=p,
                                     shape=ndata,
                                     observed=v)
        trace = pymc3.sample(3000, step=pymc3.Slice())

    pymc3.traceplot(trace)
    plt.show()

if __name__ == '__main__':
    # test_bernoulli()
    test_categorical()


import copy

import numpy as np

from archive import sampler, energy


class GibbsSampler(object):
    RM_ITERATIONS = 10
    def __init__(self, target_variables, total_energy_function, energy_gradients):
        self.target_variables = target_variables
        self.epsilon = {v:None for v in target_variables}
        self.total_energy_function = total_energy_function
        self.energy_gradients = {v:g for v,g in zip(target_variables, energy_gradients)}

    def run_sampling(self, initial_state, iterations, burn_in=0):
        samples = []
        state = initial_state
        for i in range(iterations+burn_in):
            state = self._get_next_state(state)
            vectorized_state = self._vectorize_state(state)
            if i >= burn_in:
                samples.append(vectorized_state)
        return samples

    def _get_next_state(self, current_state):
        for variable in self.target_variables:
            current_state = self._mutate_state(current_state, variable)
        return current_state

    def _mutate_state(self, current_state, variable):
        E = self._build_energy(current_state, variable)
        current_variable_value = current_state[variable]
        new_variable_value = self._sample_variable_value(current_variable_value, E, variable)
        current_state[variable] = new_variable_value
        return current_state

    def _build_energy(self, current_state, variable):
        closed_state = copy.copy(current_state)
        del closed_state[variable]
        f = self.build_arg_closure(self.total_energy_function, closed_state, variable)
        grad = self.build_arg_closure(self.energy_gradients[variable],
                                 closed_state,
                                 variable)
        return energy.Energy(eval=f, gradient=grad)


    def build_arg_closure(self, f, bound_kwargs, unbound_arg_name):
        def partial_fxn(arg):
            full_kwargs = {}
            full_kwargs.update(bound_kwargs)
            full_kwargs.update({unbound_arg_name: arg})
            return f(**full_kwargs)
        return partial_fxn


    def _sample_variable_value(self, current_variable_value, E, variable):
        nuts = sampler.NUTS()
        e = self.epsilon[variable]
        if not e:
            e = sampler.RobbinsMonroEpsilonEstimator().estimate_epsilon(E, current_variable_value, self.RM_ITERATIONS)
            self.epsilon[variable] = e
        sample = nuts.nuts_with_fixed_epsilon(current_variable_value, E, e, iterations=1)
        return sample[0]


    def _vectorize_state(self, state):
        return np.array([state[v] for v in self.target_variables])


import unittest
import functools

import numpy as np
import matplotlib.pyplot as plt

from archive import gibbs_sampler, energy
from pysterior.spikes import energy_function_spikes


class MvNormalGibbsSamplingTest(unittest.TestCase):
    def _run_gaussian_direct_sampling(self, TRUE_MU, TRUE_SIGMA):
        gaussian_fxn_spec = energy.get_bivariate_normal_spec()
        target_variables = [v.name for v in gaussian_fxn_spec.variables[0:2]]
        factory = energy.PartiallyDifferentiableFunctionFactory(gaussian_fxn_spec)
        f, f_x1 = factory.get_partial_diff('X1')
        _, f_x2 = factory.get_partial_diff('X2')
        closed_variables = {'mu': TRUE_MU, 'sigma': TRUE_SIGMA}
        closed_total_energy = functools.partial(f, **closed_variables)
        closed_gradients = [functools.partial(f_x1, **closed_variables), functools.partial(f_x2, **closed_variables)]
        sampler = gibbs_sampler.GibbsSampler(target_variables,
                                             closed_total_energy,
                                             closed_gradients)
        initial_state = {'X1': 1.0, 'X2': 2.0}
        samples = sampler.run_sampling(initial_state, iterations=1000, burn_in=20)
        return samples

    @unittest.skip('')
    def test_sampling(self):
        TRUE_MU = np.array([1,2])
        TRUE_SIGMA = np.array([[10,10],[0,10]])
        samples = self._run_gaussian_direct_sampling(TRUE_MU, TRUE_SIGMA)
        plt.plot(*zip(*samples), linewidth=0.0, marker='.')
        plt.show()

    def test_gaussian_posterior(self):
        gaussian_fxn_spec = energy.get_normal_spec()
        target_variables = [v.name for v in gaussian_fxn_spec.variables[1:]]
        data = [np.array([0,0]), np.array([1,0]), np.array([-1,0]), np.array([0,1]), np.array([0,-1])]
        E = energy_function_spikes.GaussianTotalEnergy(10000.0, 10000.0, data)
        sampler = gibbs_sampler.GibbsSampler(target_variables,
                                             E.eval_total_energy,
                                             [E.eval_energy_mu_gradient, E.eval_energy_sigma_gradient])
        initial_state = {'Mu': np.array([100,100]), 'Sigma': np.array([20, 10, 10, 15])}
        samples = sampler.run_sampling(initial_state, iterations=10, burn_in=20)
    

if __name__ == '__main__':
    unittest.main()


import os
os.path.join('..')
import unittest
from archive import energy
import theano.tensor as T

class PartiallyDiffFunctionFactoryTest(unittest.TestCase):
    def test_product(self):
        a,b = T.scalar('a'), T.scalar('b')
        product = a*b
        product_spec = energy.FunctionSpec(variables=[a,b],
                                                               output_expression=product)
        factory = energy.PartiallyDifferentiableFunctionFactory(product_spec)
        f, a_grad = factory.get_partial_diff('a')
        self.assertEqual(f(a=2, b=2), 4)
        self.assertEqual(a_grad(a=1.0, b=2.0), 2.0)

class TestClosure(unittest.TestCase):
    def test_sum_kwarg_closure(self):
        def sum_dict(a, b):
            return a+b

        add_one = energy.build_kwarg_closure(sum_dict, {'a': 1})
        self.assertEqual(add_one(b=100), 101)

    def test_sum_arg_closure(self):
        def sum_dict(a, b):
            return a+b

        add_one = energy.build_arg_closure(sum_dict, {'a': 1}, 'b')
        self.assertEqual(add_one(100), 101)

if __name__ == '__main__':
    unittest.main()


import numpy as np
from pysterior import linear_regression
from matplotlib import pyplot as plt
from sklearn.preprocessing import PolynomialFeatures
from sklearn import linear_model

def linear_regression_demo():
    TRUE_ALPHA, TRUE_SIGMA = 1, 1
    TRUE_BETA = 2.5
    size = 100
    X = np.linspace(0, 1, size)
    noise = (np.random.randn(size)*TRUE_SIGMA)
    y = (TRUE_ALPHA + TRUE_BETA*X + noise)

    lr = linear_regression.LinearRegression()
    lr.fit(X, y, 1000)
    plt.plot(X, y, linewidth=0.0, marker='x', color='g')
    pred_post_points = [lr.get_predictive_posterior_samples(x) for x in X]
    transpose = list(zip(*pred_post_points))
    for y_values in transpose:
        plt.plot(X, y_values, color='r')
    predicted_line = [lr.predict(x) for x in X]
    plt.plot(X, predicted_line)
    plt.show()

def polynomial_regression_demo():
    TRUE_ALPHA, TRUE_SIGMA = 1, 1.0
    TRUE_BETA1 = 2.5
    TRUE_BETA2 = 6.5
    TRUE_BETA3 = 2.5
    TRUE_BETA4 = -10.5
    size = 10
    X = np.linspace(-1.0, 1.0, size)
    noise = (np.random.randn(size)*TRUE_SIGMA)
    y = (TRUE_ALPHA + TRUE_BETA1*X + TRUE_BETA2*X**2 + TRUE_BETA3*X**3  + TRUE_BETA4*X**4 + noise)

    lr = linear_regression.LinearRegression()
    poly_X = PolynomialFeatures(include_bias=False, degree=4).fit_transform(X.reshape((size,1)))

    lr.fit(poly_X, y, 1000)
    pred_post_points = [lr.get_predictive_posterior_samples(x) for x in poly_X]
    transpose = list(zip(*pred_post_points))
    for y_values in transpose:
        plt.plot(X, y_values, color='r')
    predicted_line = [lr.predict(x) for x in poly_X]
    plt.plot(X, predicted_line)
    plt.plot(X, y, linewidth=0.0, marker='x', color='g')
    plt.show()

def cubic_regression_comparison():
    TRUE_ALPHA, TRUE_SIGMA = 1, 1.0
    TRUE_BETA1 = 2.5
    TRUE_BETA2 = 6.5
    TRUE_BETA3 = 2.5
    TRUE_BETA4 = -10.5
    size = 10
    X = np.linspace(-1.0, 1.0, size)
    noise = (np.random.randn(size)*TRUE_SIGMA)
    y = (TRUE_ALPHA + TRUE_BETA1*X + TRUE_BETA2*X**2 + TRUE_BETA3*X**3  + TRUE_BETA4*X**4 + noise)

    lr = linear_regression.LinearRegression()
    poly_X = PolynomialFeatures(include_bias=False, degree=4).fit_transform(X.reshape((size,1)))

    lr.fit(poly_X, y, 1000)
    pred_post_points = [lr.get_predictive_posterior_samples(x) for x in poly_X]
    transpose = list(zip(*pred_post_points))
    for y_values in transpose:
        plt.plot(X, y_values, color='r', linestyle='dotted')
    predicted_line = [lr.predict(x) for x in poly_X]
    plt.plot(X, predicted_line)
    plt.plot(X, y, linewidth=0.0, marker='x', color='g')

    ols = linear_model.LinearRegression()
    ols.fit(poly_X, y)
    predicted_line = [ols.predict(x) for x in poly_X]
    plt.plot(X, predicted_line, color='c')

    plt.plot(X, y, linewidth=0.0, marker='x', color='g')
    plt.show()

def robust_cubic_regression_comparison():
    TRUE_ALPHA, TRUE_SIGMA = 1, 1.0
    TRUE_BETA1 = 2.5
    TRUE_BETA2 = 6.5
    TRUE_BETA3 = 2.5
    TRUE_BETA4 = -10.5
    size = 12
    X = np.linspace(-1.0, 1.0, size)
    noise = (np.random.randn(size)*TRUE_SIGMA)
    y = TRUE_ALPHA + TRUE_BETA1*X + TRUE_BETA2*X**2 + TRUE_BETA3*X**3  + TRUE_BETA4*X**4 + noise
    y[3] = -10.0
    y[4] = -8.0

    lr = linear_regression.RobustLinearRegression()
    poly_X = PolynomialFeatures(include_bias=False, degree=4).fit_transform(X.reshape((size,1)))

    lr.fit(poly_X, y, 1000)
    pred_post_points = [lr.get_predictive_posterior_samples(x) for x in poly_X]
    transpose = list(zip(*pred_post_points))
    for y_values in transpose:
        plt.plot(X, y_values, color='r', linestyle='dotted')
    predicted_line = np.array([lr.predict(x) for x in poly_X])
    plt.plot(X, predicted_line, color='black')

    ols = linear_model.LinearRegression()
    ols.fit(poly_X, y)
    predicted_line = ols.predict(poly_X)
    plt.plot(X, predicted_line, color='blue')

    plt.plot(X, y, linewidth=0.0, marker='x', color='g')
    plt.show()

if __name__ == '__main__':
    # linear_regression_demo()
    # polynomial_regression_demo()
    # cubic_regression_comparison()
    robust_cubic_regression_comparison()


import unittest
import math
import numpy as np

def compose(inner_function, outer_function):
    def composed_function(*args, **kwargs):
        return outer_function(inner_function(*args, **kwargs))
    return composed_function

def subcompose_single(inner_function, target_kwarg, outer_function):
    def composed_function(**kwargs):
        inner_function_result = inner_function(kwargs[target_kwarg])
        kwargs[target_kwarg] = inner_function_result
        return outer_function(**kwargs)
    return composed_function

def sum_map(functions):
    def sum_mapped_function(*args, **kwargs):
        return sum((f(*args, **kwargs)) for f in functions)
    return sum_mapped_function

class FunctionalSpikesTest(unittest.TestCase):
    def test_composition(self):
        square = lambda x: x**2
        double = lambda x: x*2
        square_then_double = compose(square, double)
        double_then_square = compose(double, square)
        self.assertEqual(square_then_double(5), (5**2)*2)
        self.assertEqual(double_then_square(5), (5*2)**2)

    def test_subcomposition(self):
        def add_one(x):
            return x + 1

        def add(x1=None, x2=None):
            return x1 + x2

        def flatten(M):
            return M.flatten()

        add_plus_one = subcompose_single(add_one, 'x1', add)
        self.assertEqual(add(x1=2, x2=2), 4)
        self.assertEqual(add_plus_one(x1=2, x2=2), 5)
        x1_plus_sqrt_x2 = subcompose_single(math.sqrt, 'x2', add)
        self.assertEqual(x1_plus_sqrt_x2(x1=12.3, x2=4), 14.3)
        # Add flattened matrices
        flatten_x1_and_add = subcompose_single(flatten, 'x1', add)
        flatten_both_and_add = subcompose_single(flatten, 'x2', flatten_x1_and_add)
        A = np.array([[0,1],[2,3]])
        self.assertEqual(list(flatten_both_and_add(x1=A, x2=A)), [0, 2, 4, 6])

if __name__ == '__main__':
    unittest.main()


import unittest

import theano.tensor as T
from theano import function, grad
import matplotlib.pyplot as plt
from scipy.stats import shapiro

from archive import energy

from archive.sampler import NUTS
from pysterior import energy



#TODO: Make these automated tests

class TestUnivariateDistributions(unittest.TestCase):
    @unittest.skip('')
    def test_normal(self):
        TRUE_MEAN = -10.0
        TRUE_SIGMA = 3.7
        sampler = NUTS()
        X = T.scalar('X')
        output = -0.5 * ((X - TRUE_MEAN)**2) * (TRUE_SIGMA ** (-2))
        f = function([X], output, allow_input_downcast=True)
        grad_f = function([X],
                          grad(output, X),
                          allow_input_downcast=True)
        E = energy.Energy(eval=f,
                          gradient=grad_f)
        samples = sampler.nuts_with_initial_epsilon(0.0, E, iterations=10000, burn_in=100)
        w,p = shapiro(samples)
        self.assertGreater(p, 0.01)

    def test_laplace(self):
        TRUE_MEAN = -10.0
        TRUE_SCALE = 0.07
        sampler = NUTS()
        X = T.scalar('X')
        output = -1.0 * (abs(X - TRUE_MEAN) / TRUE_SCALE)
        f = function([X], output, allow_input_downcast=True)
        grad_f = function([X],
                          grad(output, X),
                          allow_input_downcast=True)
        E = energy.Energy(eval=f,
                          gradient=grad_f)
        samples = sampler.nuts_with_initial_epsilon(0.0, E, iterations=10000, burn_in=100)
        plt.hist(samples, bins=100)
        plt.show()

if __name__ == '__main__':
    unittest.main()


# This is a pilot of encapsulating pymc3 models
import unittest
import numpy as np
from linear_regression import LinearRegression, RidgeRegression, RobustLinearRegression, LassoRegression

class LinearRegressionTest(unittest.TestCase):
    def _run_multiple_linear_regression_for_model(self, model):
        np.random.seed(123)
        TRUE_ALPHA, TRUE_SIGMA = 1, 1
        TRUE_BETA = [1, 2.5]
        size = 100
        X1 = np.linspace(0, 1, size)
        X2 = np.linspace(0,.2, size)
        y = TRUE_ALPHA + TRUE_BETA[0]*X1 + TRUE_BETA[1]*X2 + np.random.randn(size)*TRUE_SIGMA
        X = np.array(list(zip(X1, X2)))
        lr = model
        lr.fit(X, y, 10000)
        samples = lr.get_samples()
        map_estimate = lr.get_map_estimate()
        expected_map = {'alpha': np.array(1.014043926179071), 'beta': np.array([ 1.46737108,  0.29347422]), 'sigma_log': np.array(0.11928775836956886)}
        self.assertAlmostEqual(float(map_estimate['alpha']), float(expected_map['alpha']), delta=1e-1)
        for true_beta, map_beta in zip(map_estimate['beta'], expected_map['beta']):
            self.assertAlmostEqual(true_beta, map_beta, delta=1e-1)
        test_point = X[7]
        true_y = TRUE_ALPHA + TRUE_BETA[0]*test_point[0] + TRUE_BETA[1]*test_point[1]
        print(true_y)
        predicted_y = lr.predict(test_point)
        print(predicted_y)
        self.assertAlmostEqual(true_y, predicted_y, delta=1e-1)

    def test_multiple_linear_regression(self):
        self._run_multiple_linear_regression_for_model(LinearRegression())

    def test_multiple_linear_regression_ridge(self):
        self._run_multiple_linear_regression_for_model(RidgeRegression(1000))

    def test_multiple_linear_regression_lasso(self):
        self._run_multiple_linear_regression_for_model(LassoRegression(1000))

    def test_multiple_linear_regression_robust(self):
        self._run_multiple_linear_regression_for_model(RobustLinearRegression())

    def _run_simple_linear_regression_for_model(self, model):
        np.random.seed(123)
        TRUE_ALPHA, TRUE_SIGMA = 1, 1
        TRUE_BETA = 2.5
        size = 100
        X = np.linspace(0, 1, size)
        noise = (np.random.randn(size)*TRUE_SIGMA)
        y = (TRUE_ALPHA + TRUE_BETA*X + noise)

        lr = model
        lr.fit(X, y, 2000)
        test_point = X[7]
        true_y = TRUE_ALPHA + TRUE_BETA*test_point
        print(true_y)
        predicted_y = lr.predict(test_point)
        print(predicted_y)
        self.assertAlmostEqual(true_y, predicted_y, delta=1e-1)


    def test_simple_linear_regression(self):
        self._run_simple_linear_regression_for_model(LinearRegression())

    def test_simple_linear_regression_ridge(self):
        self._run_simple_linear_regression_for_model(RidgeRegression(1000))

    def test_simple_linear_regression_lasso(self):
        self._run_simple_linear_regression_for_model(LassoRegression(1000))

    def test_simple_linear_regression_robust(self):
        self._run_simple_linear_regression_for_model(RobustLinearRegression())

if __name__ == '__main__':
    unittest.main()


